#include <X11/Xlib.h>
#include <X11/Xutil.h>
#include <X11/Xatom.h>
#include <GL/glx.h>
#include <cstdlib>
#include <cstring>
#include <rewindow/types/window.h>
#include <rewindow/types/cursors.h>
#include <J3ML/J3ML.hpp>
#include <rewindow/logger/logger.h>



// TODO: Move all "global" members to be instantiated class members of Window
// Doing this would break the intended "Platform-Specific" Encapsulation
// So should we do derived platform-specific subclasses?
// The intended goal of the ReWindow class is a one-stop object that handles window management on **ALL** platforms

Window window;
XEvent xev;
Display* display = XOpenDisplay(nullptr);
int defaultScreen = DefaultScreen(display);
XVisualInfo* visual;
XSetWindowAttributes xSetWindowAttributes;
XWindowAttributes windowAttributes;
Atom wmDeleteWindow;
// Make it start as floating because fuck tiling WMs
Atom windowTypeAtom;
Atom windowTypeUtilityAtom;
XSizeHints hints;
GLXContext glContext;
Cursor invisible_cursor = 0;

Vector2 render_area_position = {0, 0};
Vector2 position = {0, 0};
bool should_poll_x_for_mouse_pos = true;
typedef GLXContext (*glXCreateContextAttribsARBProc)(Display*, GLXFBConfig, GLXContext, Bool, const int*);


using namespace ReWindow;

void RWindow::Raise() const {
    Logger::Debug(std::format("Raising window '{}'", this->title));

    // Get the position of the renderable area relative to the rest of the window.
    XGetWindowAttributes(display, window, &windowAttributes);
    render_area_position =  { (float) windowAttributes.x, (float) windowAttributes.y };

    XRaiseWindow(display, window);
}

void RWindow::Lower() const
{
    Logger::Debug(std::format("Lowering window '{}'", this->title));
    XLowerWindow(display, window);
}

void RWindow::DestroyOSWindowHandle() {
    Logger::Debug(std::format("Destroying window '{}'", this->title));
    XDestroySubwindows(display, window);
    Logger::Debug(std::format("Destroyed window '{}'", this->title));
    XDestroyWindow(display, window);

    system("xset r on"); // Re-set X-server parameter to enable key-repeat.

    open = false;
}

//void RWindow::

void RWindow::SetCursorVisible(bool cursor_enable) {
    cursor_visible = cursor_enable;
    if (invisible_cursor == 0) {
        Pixmap blank_pixmap = XCreatePixmap(display, window, 1, 1, 1);
        XColor dummy; dummy.pixel = 0; dummy.red = 0; dummy.flags = 0;
        invisible_cursor = XCreatePixmapCursor(display, blank_pixmap, blank_pixmap, &dummy, &dummy, 0, 0);
        XFreePixmap(display, blank_pixmap);
    }

    if (!cursor_enable)
        XDefineCursor(display, window, invisible_cursor);

    if (cursor_enable)
        XUndefineCursor(display, window);
}

void RWindow::SetResizable(bool sizable) {
    XGetWindowAttributes(display,window,&windowAttributes);


    this->resizable = sizable;
    if (!sizable)
    {
        Logger::Debug("Once you've done this you cannot make it resizable again.");
        hints.flags = PMinSize | PMaxSize;
        hints.min_width = hints.max_width = windowAttributes.width;
        hints.min_height = hints.max_height = windowAttributes.height;
        XSetWMNormalHints(display, window, &hints);
    }
    //this->SetFlag(RWindowFlags::RESIZABLE, resizable);

}

void RWindow::SetFlag(RWindowFlags flag, bool state) {
    XGetWindowAttributes(display,window,&windowAttributes);
    flags[(int) flag] = state;
    //Once you've done this you cannot make it resizable again.
    if (flag == RWindowFlags::RESIZABLE && !state) {
        Logger::Debug("Once you've done this you cannot make it resizable again.");
        hints.flags = PMinSize | PMaxSize;
        hints.min_width = hints.max_width = windowAttributes.width;
        hints.min_height = hints.max_height = windowAttributes.height;
        XSetWMNormalHints(display, window, &hints);
    }
    Logger::Debug(std::format("Set flag '{}' to state '{}' for window '{}'", RWindowFlagToStr(flag), state, this->title));
}

void RWindow::PollEvents() {
    while(XPending(display)) {
        XNextEvent(display, &xev);

        if (xev.type == ClientMessage)
            Logger::Info(std::format("Event '{}'", "ClientMessage"));

        if (xev.xclient.message_type == XInternAtom(display, "WM_PROTOCOLS", False) && static_cast<Atom>(xev.xclient.data.l[0]) == wmDeleteWindow) {
            Close();
        }

        if (xev.type == FocusIn) {
            Logger::Debug(std::format("Event'{}'", "FocusIn"));
            XAutoRepeatOff(display);
            SetFlag(RWindowFlags::IN_FOCUS, true);

            if (!cursor_visible)
                XDefineCursor(display, window, invisible_cursor);

            // Get the position of the renderable area relative to the rest of the window.
            XGetWindowAttributes(display, window, &windowAttributes);
            render_area_position =  { (float) windowAttributes.x, (float) windowAttributes.y };
            processFocusIn();
        }

        if (xev.type == FocusOut) {
            Logger::Debug(std::format("Event '{}'", "FocusOut"));
            XAutoRepeatOn(display);

            SetFlag(RWindowFlags::IN_FOCUS, false);
            if (!cursor_visible)
                XUndefineCursor(display, window);

            processFocusOut();
        }

        if (xev.type == KeyRelease) {
            Logger::Debug(std::format("Event '{}'", "KeyRelease"));
            auto scancode = (X11Scancode) xev.xkey.keycode;
            auto key = GetKeyFromX11Scancode(scancode);
            processKeyRelease(key);
        }

        if (xev.type == KeyPress) {
            Logger::Debug(std::format("Event '{}'", "KeyPress"));
            auto scancode = (X11Scancode) xev.xkey.keycode;
            auto key = GetKeyFromX11Scancode(scancode);
            processKeyPress(key);
        }

        if (xev.type == ButtonRelease) {

            // Mouse Wheel fires both the ButtonPress and ButtonRelease instantaneously.
            // Therefore, we handle it as a specific MouseWheel event rather than a MouseButton event,
            // and only call on ButtonPress, otherwise it will appear to duplicate the mouse wheel scroll.
            if (xev.xbutton.button != 4 && xev.xbutton.button != 5) {
                MouseButton button = GetMouseButtonFromXButton(xev.xbutton.button);
                Logger::Debug(std::format("Event '{}'", "ButtonRelease"));
                processMouseRelease(button);
            }
        }

        if (xev.type == ButtonPress) {

            // Mouse Wheel fires both the ButtonPress and ButtonRelease instantaneously.
            // Therefore, we handle it as a specific MouseWheel event rather than a MouseButton event,
            // and only call on ButtonPress, otherwise it will appear to duplicate the mouse wheel scroll.
            if (xev.xbutton.button == 4) {
                processMouseWheel(-1);
            } else if (xev.xbutton.button == 5) {
                processMouseWheel(1);
            } else
            {
                MouseButton button = GetMouseButtonFromXButton(xev.xbutton.button);

                Logger::Debug(std::format("Event: MouseButtonPress {}", button.Mnemonic));

                processMousePress(button);
            }
        }

        if (xev.type == Expose)
        {
            Logger::Debug(std::format("Event '{}'", "Expose"));
        }

        // NOTE: This event is functionally useless, as it only informs of the very beginning and end of a mouse movement.
        if (xev.type == MotionNotify)
        {
            Logger::Debug(std::format("Event '{}'", "MotionNotify"));
        }

        if (xev.type == ConfigureNotify) {
            if (this->width != xev.xconfigurerequest.width || this->height != xev.xconfigurerequest.height) {
                Logger::Debug(std::format("Event '{}'", "ResizeRequest"));

                this->width = xev.xconfigurerequest.width;
                this->height = xev.xconfigurerequest.height;

                auto eventData = WindowResizeRequestEvent();
                eventData.Size = { (float) xev.xconfigurerequest.width, (float) xev.xconfigurerequest.height };
                lastKnownWindowSize = eventData.Size;

                OnResizeRequest(eventData);
                OnResizeRequestEvent(eventData);
            }

            //Window Moved.
            if (position.x != xev.xconfigurerequest.x || position.y != xev.xconfigurerequest.y)
                position = { (float) xev.xconfigurerequest.x, (float) xev.xconfigurerequest.y };
        }

    }
    previousKeyboard = currentKeyboard;
    previousMouse.Buttons = currentMouse.Buttons;
}


// Might make the window go off the screen on some window managers.
void RWindow::SetSize(int newWidth, int newHeight) {
    if (!resizable) return;

    this->width = newWidth;
    this->height = newHeight;
    XResizeWindow(display, window, newWidth, newHeight);
    XFlush(display);
    Logger::Info(std::format("Set size for '{}' to {} x {}", this->title, newWidth, newHeight));
}

Vector2 RWindow::GetAccurateMouseCoordinates() const {

    Window root_return, child_return;
    int root_x_ret, root_y_ret;
    int win_x_ret, win_y_ret;
    uint32_t mask_return;

    // This seems to be relative to the top left corner of the renderable area.
    bool mouseAvailable = XQueryPointer(display, window, &root_return, &child_return, &root_x_ret, &root_y_ret, &win_x_ret, &win_y_ret, &mask_return);

    if (mouseAvailable) {
        // TODO: normalize coordinates from displaySpace to windowSpace
        // TODO: fire mouse movement event
        Vector2 m_coords = { (float) win_x_ret, (float) win_y_ret };
        return m_coords;
    }
    return Vector2::Zero;
}


Vector2 RWindow::GetSize() const {
    return {(float) this->width, (float) this->height};
}

//Vector2 RWindow::getLastKnownResize() const
//{
//    return lastKnownWindowSize;
//}

// TODO: implement integer vector2/3 types
Vector2 RWindow::GetPos() const {
    return position;
}

void RWindow::SetPos(int x, int y) {
    XMoveWindow(display, window, x, y);
    position = { (float) x, (float) y };
}

void RWindow::SetPos(const Vector2& pos) {
    SetPos(pos.x, pos.y);
}


void RWindow::GLSwapBuffers() {
    glXSwapBuffers(display,window);
}


void RWindow::Fullscreen() {
    Logger::Info(std::format("Fullscreening '{}'", this->title));
    fullscreen_mode = true;

    XEvent xev;
    Atom wm_state   = XInternAtom(display, "_NET_WM_STATE", true);
    Atom wm_fullscreen = XInternAtom(display, "_NET_WM_STATE_FULLSCREEN", true);

    XChangeProperty(display, window, wm_state, XA_ATOM, 32, PropModeReplace, (unsigned char *)&wm_fullscreen, 1);
    memset(&xev, 0, sizeof(xev));
    xev.type = ClientMessage;
    xev.xclient.window = window;
    xev.xclient.message_type = wm_state;
    xev.xclient.format = 32;
    xev.xclient.data.l[0] = 1; // _NET_WM_STATE_ADD
    xev.xclient.data.l[1] = fullscreen_mode;
    xev.xclient.data.l[2] = 0;
    XSendEvent(display, DefaultRootWindow(display), False, SubstructureNotifyMask | SubstructureRedirectMask, &xev);
    Logger::Debug(std::format("Fullscreened '{}'", this->title));
}

void RWindow::RestoreFromFullscreen() {
    Logger::Debug(std::format("Restoring '{}' from Fullscreen", this->title));
    fullscreen_mode = false;
    XEvent xev;
    Atom wm_state = XInternAtom(display, "_NET_WM_STATE", False);
    Atom fullscreen = XInternAtom(display, "_NET_WM_STATE_FULLSCREEN", False);
    memset(&xev, 0, sizeof(xev));
    xev.type = ClientMessage;
    xev.xclient.window = window;
    xev.xclient.message_type = wm_state;
    xev.xclient.format = 32;
    xev.xclient.data.l[0] = 0; // _NET_WM_STATE_REMOVE
    xev.xclient.data.l[1] = fullscreen_mode;
    xev.xclient.data.l[2] = 0;
    XSendEvent(display, DefaultRootWindow(display), False, SubstructureNotifyMask | SubstructureRedirectMask, &xev);
    Logger::Debug(std::format("Restored '{}' from Fullscreen", this->title));
}

void RWindow::SetVsyncEnabled(bool b) {
    PFNGLXSWAPINTERVALEXTPROC glXSwapIntervalEXT;
    glXSwapIntervalEXT = (PFNGLXSWAPINTERVALEXTPROC)glXGetProcAddressARB((const GLubyte*)"glXSwapIntervalEXT");
    glXSwapIntervalEXT(display, window, b);
}


void RWindow::SetCursorStyle(CursorStyle style) const {
    u32 x11_cursor_resolved_enum = static_cast<u32>(style.X11Cursor);
    Cursor c = XCreateFontCursor(display, x11_cursor_resolved_enum);
    XDefineCursor(display, window, c);
}

void RWindow::Open() {
    xSetWindowAttributes.border_pixel = BlackPixel(display, defaultScreen);
    xSetWindowAttributes.background_pixel = BlackPixel(display, defaultScreen);
    xSetWindowAttributes.override_redirect = True;
    xSetWindowAttributes.event_mask = ExposureMask;

    if (renderer == RenderingAPI::OPENGL) {
        auto glXCreateContextAttribsARB = (glXCreateContextAttribsARBProc)glXGetProcAddressARB((const GLubyte*)"glXCreateContextAttribsARB");
        XVisualInfo* vi = nullptr;

        // Fallback to the old way if you somehow don't have this.
        if (!glXCreateContextAttribsARB) {
            GLint glAttributes[] = {GLX_RGBA, GLX_DEPTH_SIZE, 24, GLX_DOUBLEBUFFER, None};
            visual = glXChooseVisual(display, defaultScreen, glAttributes);
            glContext = glXCreateContext(display, visual, nullptr, GL_TRUE);
            window = XCreateWindow(display, RootWindow(display, defaultScreen), 0, 0, width, height, 0, visual->depth,
                                   InputOutput, visual->visual, CWBackPixel | CWColormap | CWBorderPixel | NoEventMask, &xSetWindowAttributes);

            ReWindow::Logger::Debug("Created OpenGL Context with glXCreateContext.");
        }
        else {
            static int visual_attributes[]
                    {
                    GLX_X_RENDERABLE, True, GLX_DRAWABLE_TYPE, GLX_WINDOW_BIT,
                    GLX_RENDER_TYPE, GLX_RGBA_BIT, GLX_X_VISUAL_TYPE, GLX_TRUE_COLOR,
                    GLX_RED_SIZE, 8, GLX_GREEN_SIZE, 8, GLX_BLUE_SIZE, 8,
                    GLX_ALPHA_SIZE, 8, GLX_DEPTH_SIZE, 24, GLX_STENCIL_SIZE, 8,
                    GLX_DOUBLEBUFFER, True, None
                    };

            int fb_count;
            GLXFBConfig *fb_configurations = glXChooseFBConfig(display, defaultScreen, visual_attributes, &fb_count);
            if (!fb_configurations || fb_count == 0)
                throw std::runtime_error("Couldn't get framebuffer configuration.");

            GLXFBConfig best_fbc = fb_configurations[0];
            vi = glXGetVisualFromFBConfig(display, best_fbc);

            if (!vi)
                throw std::runtime_error("Couldn't visual info from framebuffer configuration.");

            xSetWindowAttributes.colormap = XCreateColormap(display, RootWindow(display, vi->screen), vi->visual, AllocNone);
            window = XCreateWindow(display, RootWindow(display, vi->screen), 0, 0, width, height, 0, vi->depth, InputOutput,
                                   vi->visual, CWBackPixel | CWColormap | CWBorderPixel, &xSetWindowAttributes);

            // TODO allow the user to specify what OpenGL version they want.
            int context_attributes[] { GLX_CONTEXT_MAJOR_VERSION_ARB, 2, GLX_CONTEXT_MINOR_VERSION_ARB, 1, None };
            glContext = glXCreateContextAttribsARB(display, best_fbc, nullptr, True, context_attributes);
            XFree(fb_configurations);
        }

        if (!glContext)
            throw std::runtime_error("Couldn't create the OpenGL context.");
        if (!glXMakeCurrent(display, window, glContext))
            throw std::runtime_error("Couldn't change OpenGL context to current.");
        if (vi)
            XFree(vi);

        ReWindow::Logger::Debug("Created OpenGL Context with glXCreateContextAttribsARB");
    }

    XSelectInput(display, window, ExposureMask | KeyPressMask | KeyReleaseMask | ButtonPressMask | ButtonReleaseMask | PointerMotionMask |
                                  PointerMotionHintMask | FocusChangeMask | StructureNotifyMask | SubstructureRedirectMask | SubstructureNotifyMask | CWColormap );
    XMapWindow(display, window);
    XStoreName(display, window, title.c_str());

    windowTypeAtom = XInternAtom(display, "_NET_WM_WINDOW_TYPE", False);
    windowTypeUtilityAtom = XInternAtom(display, "_NET_WM_WINDOW_TYPE_UTILITY", False);
    XChangeProperty(display, window, windowTypeAtom, XA_ATOM, 32, PropModeReplace, (unsigned char*)&windowTypeUtilityAtom, 1);

    wmDeleteWindow = XInternAtom(display, "WM_DELETE_WINDOW", False);
    XSetWMProtocols(display, window, &wmDeleteWindow, 1);

    XGetWindowAttributes(display, window, &windowAttributes);
    render_area_position = {(float)windowAttributes.x, (float)windowAttributes.y};

    open = true;
    processOnOpen();
}

void RWindow::SetTitle(const std::string &title) {
    this->title = title;
    XStoreName(display, window, title.c_str());
}



Vector2 RWindow::GetPositionOfRenderableArea() const {
    return render_area_position;
}

std::string RWindow::getGraphicsDriverVendor() {
    return std::string(reinterpret_cast<const char*>(glGetString(GL_VENDOR)));
}


// TODO: Implement ControllerButton map

